import attr
from typing import Mapping, Any, Set, Tuple
from pyrsistent import pmap
from pyrsistent.typing import PVector, PMap, PSet
from .gates import GateDef, _qubit_ids
from .helpers import map_seq_to_seq
from inspect import signature
from fractions import Fraction
import numpy


@attr.s(frozen=True)
class Instruction(object):
    """
    An instruction: a gate definition paired with parameter bindings
    and bit bindings
    """
    gate_def = attr.ib(type=GateDef)
    bit_bindings = attr.ib(type=PVector[int], converter=_qubit_ids)
    parameter_bindings = attr.ib(type=PMap[str, Any],
                                 default=pmap(),
                                 converter=pmap)
    # 'metadata' contains extra information not necessarily
    # understood by translations, such as "classical bit" arguments,
    # etc. which are not normally "part" of a quantum circuit.
    metadata = attr.ib(type=PMap[str, Any],
                       default=pmap(),
                       converter=pmap)

    @parameter_bindings.validator
    def check_parameter_bindings(self, attribute, value):
        if not set(value.keys()).issubset(self.gate_def.parameter_names):
            raise ValueError(
                "parameter bindings must bind parameters in the gate def")

    @bit_bindings.validator
    def check_bit_bindings(self, attribute, value):
        if len(value) != len(self.gate_def.qubit_ids):
            raise ValueError(
                "number of bit bindings must be equal to #bits in the gate def"
            )

    def __str__(self):
        parameter_bindings_str = ",".join(
            [f"{k}={v}" for k, v in self.parameter_bindings.items()])
        bit_bindings_str = ",".join([f"{x}" for x in self.bit_bindings])
        metadata = "" if len(self.metadata) == 0 else ", (" + ",".join(["{k}={v}" for k,v in self.metadata.items()]) + ")"
        return f"{self.gate_def.name}({parameter_bindings_str}), ({bit_bindings_str}){metadata})"


def instruction_parameters_are_fully_bound(i: Instruction) -> bool:
    """
    Whether or not all parameters in the instruction are fully
    bound to values, not functions.  This is important in deciding if a 
    circuit is fully bound (a target for a match pattern)
    or only a pattern (only a subset of 
    parameters is bound), or a match target (some parameters are bound
    to callables)
    """
    return all([(k in i.parameter_bindings)
                for k in i.gate_def.parameter_names])


def _is_valid_executable_parameter_value(v: Any) -> bool:
    """
    whether the instruction parameter is valid for execution (is numeric)
    """
    return numpy.isscalar(v)


def _is_valid_replacement_parameter_value(v: Any) -> bool:
    """
    Determines if a given value is a valid replacement
    parameter, ie if it is a tuple, callable of one argument,
    or numeric parameter
    """
    return ((isinstance(v, tuple) and isinstance(v[0], int)
             and isinstance(v[1], str))
            or (callable(v) and len(signature(v).parameters) == 1)
            or (numpy.isscalar(v)))


def audit_instruction_for_executable(i: Instruction) -> PMap:
    """
    Here to give more information about why an instruction isn't executable
    """
    result = {}
    if not instruction_parameters_are_fully_bound(i):
        result['instruction_parameters_not_fully_bound'] = True
    invalid_bindings = {
        k: (v, type(v))
        for k, v in i.parameter_bindings.items()
        if not _is_valid_executable_parameter_value(v)
    }
    if len(invalid_bindings) > 0:
        result['invalid parameter bindings'] = invalid_bindings # type: ignore
    return pmap(result)


def instruction_is_valid_executable(i: Instruction) -> bool:
    """
    An instruction is a valid executable instruction if it is
    fully bound and if all parameter values are concrete
    numbers
    """
    return (instruction_parameters_are_fully_bound(i) and all([
        _is_valid_executable_parameter_value(v)
        for v in i.parameter_bindings.values()
    ]))


def instruction_is_valid_replacement(i: Instruction) -> bool:
    """
    An instruction is a valid replacement instruction if it is
    fully bound and if all parameter "values" are either a
    value, tuple, or callable of one argument
    """
    return (instruction_parameters_are_fully_bound(i) and all([
        _is_valid_replacement_parameter_value(v)
        for v in i.parameter_bindings.values()
    ]))


def instruction_parameter_bindings_match(pattern: Instruction,
                                         target: Instruction) -> bool:
    """
    the parameters of an instruction "pattern" matches a target if all
    bound parameters in the pattern have keys present in the target
    and the bound values are the same
    """
    return set(pattern.parameter_bindings.keys()).issubset(
        set(target.parameter_bindings.keys())) \
        and all([pattern.parameter_bindings[x]
               == target.parameter_bindings[x]
               for x in pattern.parameter_bindings.keys()])


def instruction_pattern_matches_target(pattern: Instruction,
                                       target: Instruction) -> bool:
    """
    An instruction pattern matches a target if it has the same gate name,
    and matching parameter bindings
    """
    return ((pattern.gate_def.name == target.gate_def.name)
            and instruction_parameter_bindings_match(pattern, target))


def instruction_bit_bindings_map(
        instruction: Instruction) -> PMap[int, PSet[int]]:
    """
    Returns a "binding map" of bit ids to bit assignments;
    in other words, an instruction binding the gate CX with
    bit ids 0 and 1 to circuit bits 7 and 8 would return
    the map {0:7, 1:8}
    """
    qubit_ids = instruction.gate_def.qubit_ids
    bit_assignments = instruction.bit_bindings
    return map_seq_to_seq(qubit_ids, bit_assignments)


def _remapped_parameter(parameter_map: Mapping[Tuple[int, str], Any],
                        parameter_value: Any):
    """
    Remaps a parameter in a match target.  There are three
    options for a match target value:

    * If the value is a callable, call it, passing the parameter
      map as the argument
    * If the value is a tuple, replace it with the appropriate
      value from the parameter map
    * Otherwise, just keep the original value
    """
    if callable(parameter_value):
        result = parameter_value(parameter_map)
    elif isinstance(parameter_value, tuple):
        # ignoring type here as we *should* have a
        # value of type (int, str)->number
        result = parameter_map[parameter_value]  # type: ignore
    else:
        result = parameter_value
    return result


def remapped_instruction(qubit_map: Mapping[int, int],
                         parameter_map: Mapping[Tuple[int, str], Any],
                         target: Instruction) -> Instruction:
    """
    This remaps an instruction given a new qubit mapping (from
    target qubits in one circuit to target qubits in another) and
    a new parameter map (from the parameters in the original circuit,
    ie the keys in the parameter map are tuples of (index, parameter_name)

    """
    new_parameters = {
        # ignoring type as the following seems to confuse it; v can
        # either be a key into the parameter map or a tuple
        k: _remapped_parameter(parameter_map, v)
        for k, v in target.parameter_bindings.items()
    }
    new_bit_bindings = [qubit_map[b] for b in target.bit_bindings]
    return Instruction(
        gate_def=target.gate_def,
        parameter_bindings=new_parameters,  # type: ignore
        bit_bindings=new_bit_bindings)



"""
Pyzx is a toolkit for manipulating circuits using the ZX
calculus.
"""

from .pyzx_dialect import (dialect, native_to_ir, ir_to_native,
                             native_circuits_are_equivalent, audit)



import qiskit  # type: ignore
from qcware_transpile.gates import GateDef, Dialect
from qcware_transpile.circuits import Circuit
from qcware_transpile.instructions import Instruction
from qcware_transpile.helpers import map_seq_to_seq_unique
from pyrsistent import pset, pmap, pvector
from pyrsistent.typing import PSet, PMap
from typing import Tuple, Any, Set, Generator, List, Dict
from inspect import isclass, signature
import numpy as np  # type: ignore
from functools import lru_cache
from itertools import accumulate
from icontract import require

__dialect_name__ = "qiskit"


@lru_cache(1)
def qiskit_gatethings() -> PSet[Any]:
    """
    The set of all things in qiskit which represent a gate
    """
    possible_things = qiskit.circuit.library.__dict__.values()
    return pset({
        x
        for x in possible_things
        if isclass(x) and issubclass(x, qiskit.circuit.Instruction)
    })


@lru_cache(1)
def name_to_class() -> PMap[str, type]:
    return pmap({gatething_name(c): c for c in attempt_gatedefs()[2]})


@lru_cache(1)
def class_to_name() -> PMap[type, str]:
    """
    Returns a map of classes to qiskit names
    """
    return pmap({c: gatething_name(c) for c in attempt_gatedefs()[2]})


def parameter_names_from_gatething(thing: qiskit.circuit.Gate) -> PSet[str]:
    sig = signature(thing.__init__)
    result = set(sig.parameters.keys()).difference(
        {'self', 'label', 'ctrl_state'})
    return pset(result)


def number_of_qubits_from_gatething(thing: qiskit.circuit.Gate) -> int:
    # unfortunately it's not obvious from just the class how many qubits
    # the gate can operate on, and pretty much the only path forward
    # seems to be to instantiate the gate and see.  That means coming
    # up with fake parameter arguments.
    param_names = parameter_names_from_gatething(thing)
    # Obviously not every parameter can be a float, but let's start
    # there.  setting everything to zero also takes care of the
    # ctrl_state parameter which doesn't matter so much for basic gate
    # instantiation, but could affect specialized applications
    # also note that the standard for qiskit is that the first X bits of
    # a controlled gate are the control bits
    # ("In the new gate the first ``num_ctrl_qubits``
    # of the gate are the controls."), see
    # https://github.com/Qiskit/qiskit-terra/qiskit/circuit/controlledgate.py
    params = {k: 0 for k in param_names}
    g = thing(**params)
    return g.num_qubits


def gatething_name(thing: type) -> str:
    """Attempts to get the qiskit "name" from a qiskit gate class.  This
    is a mild irritation-- qiskit does a few things like lists of basis
    gates and transpilation targets through the "name" of a gate
    rather than the gate class.  That's fine, but the problem is you
    have to create an instance of the gate to get its "name"
    """
    parameter_names = parameter_names_from_gatething(thing)
    # we're so far assuming it's ok to call gate instantiation with
    # all parameters being the float 0
    parameters = {name: 0 for name in parameter_names}
    gate = thing(**parameters)
    return gate.name


def gatedef_from_gatething(thing) -> GateDef:
    return GateDef(
        name=gatething_name(thing),
        parameter_names=parameter_names_from_gatething(thing),  # type: ignore
        qubit_ids=number_of_qubits_from_gatething(thing))


# some gates are problematic -- in particular Qiskit's "gates" which
# really just generate other gates, for example those which take a
# number of bits as an argument.  for now we are disabling those.
# we also disable some gates which resolve to identical "names"
# such as C3XGate and C4XGate
Problematic_gatenames = pset({"ms", "mcx", "barrier"})


@lru_cache(1)
def attempt_gatedefs() -> Tuple[PSet[GateDef], PSet[str], PSet[type]]:
    """
    Iterate through all the gate definitions and build a set of successful
    gate definitions and a list of things which could not be converted
    """
    all_things = qiskit_gatethings()
    success = set()
    successful_things = set()
    failure = set()
    for thing in all_things:
        try:
            gd = gatedef_from_gatething(thing)
            if gd.name in Problematic_gatenames:
                failure.add(thing.__name__)
            else:
                success.add(gd)
                successful_things.add(thing)
        except Exception:
            failure.add(thing.__name__)
    return (pset(success), pset(failure), pset(successful_things))


def gate_defs() -> PSet[GateDef]:
    return attempt_gatedefs()[0]


def gate_names() -> PSet[str]:
    return pset([g.name for g in gate_defs()])


@lru_cache(1)
def dialect() -> Dialect:
    """
    The qiskit dialect
    """
    return Dialect(name=__dialect_name__,
                   gate_defs=gate_defs())  # type: ignore


@lru_cache(1)
def valid_gatenames() -> Set[str]:
    return {g.name for g in dialect().gate_defs}


def parameter_bindings_from_gate(gate: qiskit.circuit.Gate) -> PMap[str, Any]:
    # sometimes transpile makes ParameterExpressions and here we try to
    # coerce those into floats... which may not always be the right idea
    values = [
        x
        if not isinstance(x, qiskit.circuit.ParameterExpression) else float(x)
        for x in gate.params
    ]
    names = [k for k, v in signature(gate.__init__).parameters.items()
             ][0:len(values)]
    return map_seq_to_seq_unique(names, values)


def native_instructions(
    qc: qiskit.QuantumCircuit
) -> Generator[Tuple[qiskit.circuit.Gate, List[qiskit.circuit.Qubit],
                     List[qiskit.circuit.Clbit]], None, None]:
    """
    Iterates over the circuit.  Does *NOT* reverse the circuit beforehand,
    because that elides the reversed qregs for mapping.  We just skip
    barriers, because they are problematic (they are multi-qubit, which
    we don't handle well yet, and they are unsupported by any other kit).
    """
    gates_to_skip = {'barrier'}
    for (instruction, qubits, clbits) in qc.data:
        if instruction.name not in gates_to_skip:
            yield instruction, qubits, clbits


@require(lambda qubit, qubits: qubit in qubits)
def raw_qubit_index(qubit: qiskit.circuit.Qubit,
                    qubits: List[qiskit.circuit.Qubit]) -> int:
    """
    Given a qubit object and a list of quantum registers, 
    calculates the "Raw qubit index"
    """
    return qubits.index(qubit)


@require(lambda clbit, clbits: clbit in clbits)
def raw_clbit_index(clbit: qiskit.circuit.Clbit,
                    clbits: List[qiskit.circuit.Clbit]) -> int:
    """
    Given a qubit object and a list of quantum registers, 
    calculates the "Raw qubit index"
    """
    return clbits.index(clbit)


def clbits(circuit: Circuit):
    """
    A sorted list of the classical bit ids in the IR representation; used
    for reconstruction of a qiskit circuit to/from IR, with difficulties
    """
    return list(
        sorted(set().union(
            *[x.metadata.get('clbits', []) for x in circuit.instructions])))


@require(lambda gate: gate.name in valid_gatenames())
def ir_instruction_from_native(
        gate: qiskit.circuit.Gate, qubits: List[qiskit.circuit.Qubit],
        clbits: List[qiskit.circuit.Clbit],
        circuit_qubits: List[qiskit.circuit.Qubit],
        circuit_clbits: List[qiskit.circuit.Clbit]) -> Instruction:
    return Instruction(
        gate_def=gatedef_from_gatething(gate.__class__),
        parameter_bindings=parameter_bindings_from_gate(gate),  # type: ignore
        bit_bindings=[raw_qubit_index(qb, circuit_qubits) for qb in qubits],
        # this below must be a pvector to handle some hashing
        metadata={
            'clbits':
            pvector([raw_clbit_index(cb, circuit_clbits) for cb in clbits])
        } if len(clbits) > 0 else {})  # type: ignore


def native_to_ir(qc: qiskit.QuantumCircuit) -> Circuit:
    """
    Return a transpile-style Circuit object from a qiskit Circuit object
    """
    rqc = qc.reverse_bits()
    instructions = list(
        ir_instruction_from_native(x[0], x[1], x[2], rqc.qubits, rqc.clbits)
        for x in native_instructions(rqc))
    qubits = list(range(qc.num_qubits))
    return Circuit(
        dialect_name=__dialect_name__,
        instructions=instructions,  # type: ignore
        qubits=qubits)  # type: ignore


def qiskit_gate_from_instruction(i: Instruction):
    """
    Create a qiskit Gate object from an instruction
    """
    gclass = name_to_class()[i.gate_def.name]
    gate = gclass(**i.parameter_bindings)
    return gate


def ir_to_native(c: Circuit) -> qiskit.QuantumCircuit:
    """
    Make a qiskit circuit from a qcware_transpile Circuit
    """
    # qiskit wants the number of qubits first.
    num_qubits = max(c.qubits) - min(c.qubits) + 1
    _clbits = clbits(c)
    # this is slightly different because we only list classical bits used and assume
    # they start from 0 without checking for unused edge bits.
    num_clbits = 0 if len(
        _clbits) == 0 else max(_clbits) + 1  #  - min(_clbits) + 1
    result = qiskit.QuantumCircuit(
        num_qubits) if num_clbits == 0 else qiskit.QuantumCircuit(
            num_qubits, num_clbits)
    for instruction in c.instructions:
        g = qiskit_gate_from_instruction(instruction)
        if 'clbits' in instruction.metadata:
            result.append(g, instruction.bit_bindings,
                          instruction.metadata['clbits'])
        else:
            result.append(g, instruction.bit_bindings)
    result = result.reverse_bits()
    return result


def native_circuits_are_equivalent(c1: qiskit.QuantumCircuit,
                                   c2: qiskit.QuantumCircuit) -> bool:
    """
    Whether or not two circuits are equivalent.  Not having a test_equivalence
    method here, we brute-force it by evaluating statevectors
    """
    backend = qiskit.Aer.get_backend('statevector_simulator')
    sv1 = qiskit.execute(c1, backend).result().data()['statevector']
    sv2 = qiskit.execute(c2, backend).result().data()['statevector']
    return np.isclose(sv1, sv2).all()


def audit(c: qiskit.QuantumCircuit) -> Dict:
    """
    Retrieve a dictionary with various members indicating
    any aspects of the circuit which would make it not
    convertible to the IR
    """
    # check for classical instructions
    # unhandled_classical_instructions = set()
    # rqc = c.reverse_bits()
    # for (instruction, qubits, cbits) in rqc.data:
    #     if (len(cbits) != 0):
    #         unhandled_classical_instructions.add(
    #             instruction.__class__.__name__)

    invalid_gate_names = set()
    for g, qubits, clbits in native_instructions(c):
        if g.name not in valid_gatenames():
            invalid_gate_names.add(gatething_name(g))
    result = {}
    if len(invalid_gate_names) > 0:
        result['invalid_gate_names'] = invalid_gate_names

    # if len(unhandled_classical_instructions) != 0:
    #     result['unhandled_classical_instructions'] = \
    #         unhandled_classical_instructions
    return result



import pyzx  # type: ignore
from qcware_transpile.gates import GateDef, Dialect
from qcware_transpile.circuits import Circuit
from qcware_transpile.instructions import Instruction
from pyrsistent import pset, pmap
from pyrsistent.typing import PSet, PMap
from typing import Tuple, Any, Set, Sequence, Generator, Dict
from inspect import isclass, signature
from functools import lru_cache
from icontract import require

__dialect_name__ = "pyzx"

# these gates don't translate yet as we don't have the concept of
# variable-qubit gates or no-qubit instructions
_do_not_include_instructions = {"ParityPhase", "InitAncilla", "PostSelect"}


@lru_cache(1)
def pyzx_gatethings() -> PSet[Any]:
    """
    The set of all things in qiskit which represent a gate.  We
    don't include the ParityPhase gate because it handles variable
    numbers of qubits.
    """
    possible_things = [
        x for x in pyzx.circuit.gates.__dict__.values()
        if isclass(x) and issubclass(x, pyzx.circuit.Gate) and x !=
        pyzx.circuit.Gate and x.__name__ not in _do_not_include_instructions
    ]
    return pset(possible_things)


def parameter_names_from_gatething(thing: pyzx.circuit.Gate) -> PSet[str]:
    sig = signature(thing.__init__)
    result = set(sig.parameters.keys()).difference(
        {'self', 'target', 'control', 'ctrl1', 'ctrl2', 'label'})
    return pset(result)


def number_of_qubits_from_gatething(thing: pyzx.circuit.Gate) -> int:
    # unfortunately it's not obvious from just the class how many qubits
    # the gate can operate on, and pretty much the only path forward
    # seems to be to instantiate the gate and see.  That means coming
    # up with fake parameter arguments.
    param_names = list(signature(thing.__init__).parameters.keys())
    # in pyzx rather than having a vector of target bits, you have bit names,
    # such as "ctrl1", "ctrl2", "target"
    result = 0
    if "target" in param_names:
        result = result + 1
    if "control" in param_names:
        result = result + 1
    else:
        if "ctrl1" in param_names:
            result = result + 1
        if "ctrl2" in param_names:
            result = result + 1
    return result


def gatedef_from_gatething(thing) -> GateDef:
    return GateDef(
        name=thing.__name__,
        parameter_names=parameter_names_from_gatething(thing),  # type: ignore
        qubit_ids=number_of_qubits_from_gatething(thing))


# some gates are problematic -- in particular Qiskit's "gates" which
# really just generate other gates, for example those which take a
# number of bits as an argument.  for now we are disabling those
Problematic_gatenames: PSet[str] = pset({})


def attempt_gatedefs() -> Tuple[PSet[GateDef], PSet[str]]:
    """
    Iterate through all the gate definitions and build a set of successful
    gate definitions and a list of things which could not be converted
    """
    all_things = pyzx_gatethings()
    success = set()
    failure = set()
    for thing in all_things:
        try:
            gd = gatedef_from_gatething(thing)
            if gd.name in Problematic_gatenames:
                failure.add(thing.__name__)
            else:
                success.add(gd)
        except Exception:
            failure.add(thing.__name__)
    return (pset(success), pset(failure))


def gate_defs() -> PSet[GateDef]:
    s, f = attempt_gatedefs()
    return s


def gate_names() -> PSet[str]:
    return pset([g.name for g in gate_defs()])


@lru_cache(1)
def dialect() -> Dialect:
    """
    The pyzx dialect
    """
    return Dialect(name=__dialect_name__,
                   gate_defs=gate_defs())  # type: ignore


@lru_cache(1)
def valid_gatenames() -> Set[str]:
    return {g.name for g in dialect().gate_defs}


def possible_parameter_names():
    return set().union(*[x.parameter_names for x in dialect().gate_defs])


def parameter_bindings_from_gate(gate: pyzx.circuit.Gate) -> PMap[str, Any]:
    # this is a little trickier than others.  Rather than storing parameters
    # in a dictionary, pyzx stores them as individual Gate members, ie
    # self.target, self.phi, etc.
    result = {
        str(k): getattr(gate, k)
        for k in possible_parameter_names()
        if k in signature(gate.__init__).parameters.keys()
    }
    return pmap(result)


def qubit_bindings(g: pyzx.circuit.Gate):
    """
    Gets an ordered set of qubit bindings for the given gate.
    Pyzx, rather sensibly, does gates with names bound to qubits,
    whereas we usually store an ordered list of qubits.  For pyzx,
    then,
    1 qubit -> target
    2 qubit -> control, target
    3 qubit -> ctrl1, ctrl2, target
    """
    if "ctrl1" in dir(g):
        result = [g.ctrl1, g.ctrl2, g.target]
    elif "control" in dir(g):
        result = [g.control, g.target]
    else:
        result = [g.target]
    return result


def native_instructions(
        pc: pyzx.circuit.Circuit) -> Generator[pyzx.circuit.Gate, None, None]:
    for g in pc.gates:
        yield g


@require(lambda g: g.__class__.__name__ in valid_gatenames())
def ir_instruction_from_native(g: pyzx.circuit.Gate) -> Instruction:
    return Instruction(
        gate_def=gatedef_from_gatething(g.__class__),
        parameter_bindings=parameter_bindings_from_gate(g),  # type: ignore
        bit_bindings=qubit_bindings(g))


def native_to_ir(pc: pyzx.circuit.Circuit) -> Circuit:
    """
    Return a transpile-style Circuit object from a qiskit Circuit object
    """
    instructions = list(
        (ir_instruction_from_native(x) for x in native_instructions(pc)))
    qubits = list(range(pc.qubits))
    return Circuit(
        dialect_name=__dialect_name__,
        instructions=instructions,  # type: ignore
        qubits=qubits)  # type: ignore


def pyzx_qubit_bindings(qubits: Sequence[int]):
    if len(qubits) == 3:
        result = {'ctrl1': qubits[0], 'ctrl2': qubits[1], 'target': qubits[2]}
    elif len(qubits) == 2:
        result = {'control': qubits[0], 'target': qubits[1]}
    elif len(qubits) == 1:
        result = {'target': qubits[0]}
    else:
        raise ValueError(f"invalid number of qubits: {qubits}")
    return result


def pyzx_gate_from_instruction(i: Instruction):
    """
    Create a pyzx Gate object from an instruction.  Unlike
    some other toolkits, a "gate" in pyzx is a fully-
    instantiated instruction (with qubit assignments)
    """
    gclass = getattr(pyzx.circuit.gates, i.gate_def.name)
    parms = i.parameter_bindings
    parms = parms.update(pyzx_qubit_bindings(i.bit_bindings))
    gate = gclass(**parms)
    return gate


def ir_to_native(c: Circuit) -> pyzx.Circuit:
    """
    Make a pyzx circuit from a qcware_transpile Circuit
    """
    # qiskit wants the number of qubits first.
    num_qubits = max(c.qubits) - min(c.qubits) + 1
    result = pyzx.Circuit(num_qubits)
    for instruction in c.instructions:
        g = pyzx_gate_from_instruction(instruction)
        result.add_gate(g)
    return result


def native_circuits_are_equivalent(c1: pyzx.Circuit, c2: pyzx.Circuit) -> bool:
    """
    Whether or not two circuits are equivalent.  This is used to check
    for equivalence of the sequence of instructions, for testing; while
    pyzx has a more complicated test for equivalence, we here check only
    that the series of gates are the same
    """
    return c1.gates == c2.gates  # c1.verify_equality(c2) fails with FSim gates


def audit(c: pyzx.Circuit) -> Dict:
    """
    Retrieve a dictionary with various members indicating
    any aspects of the circuit which would make it not
    convertible to the IR
    """
    invalid_gate_names = set()
    for g in native_instructions(c):
        if g.__class__.__name__ not in valid_gatenames():
            invalid_gate_names.add(g.__class__.__name__)
    result = {}
    if len(invalid_gate_names) > 0:
        result['invalid_gate_names'] = invalid_gate_names
    return result



import attr
import itertools
from typing import Set, Tuple, Sequence, Mapping, Any, Optional
from pyrsistent import pvector, pmap, pset
from pyrsistent.typing import PMap, PSet, PVector
from icontract import require  # type: ignore
from .helpers import reverse_map
from .gates import Dialect
from .instructions import (Instruction, instruction_bit_bindings_map,
                           instruction_parameters_are_fully_bound,
                           instruction_is_valid_executable,
                           instruction_is_valid_replacement,
                           instruction_pattern_matches_target)


@attr.s(frozen=True)
class Circuit(object):
    dialect_name = attr.ib(type=str)
    instructions = attr.ib(type=PVector[Instruction], converter=pvector)
    qubits = attr.ib(type=PSet[Any], converter=pset)

    @classmethod
    def from_tuples(cls, dialect: Dialect,
                    instructions: Sequence[Tuple[str, Mapping,
                                                 Sequence[int]]]):
        """
        A simplified version of from_instructions, where instead of
        specifying the actual instruction objects, you specify
        just the gate names, dict of parameters, and sequence of bit bindings
        """
        real_instructions = [
            Instruction(gate_def=dialect.gate_named(x[0]),
                        parameter_bindings=x[1],
                        bit_bindings=x[2]) for x in instructions
        ]
        return Circuit.from_instructions(dialect.name, real_instructions)

    @classmethod
    def from_instructions(cls,
                          dialect_name: str,
                          instructions: Sequence[Instruction],
                          qubits: Optional[Set[Any]] = None):
        if qubits is None:
            new_qubits: PSet = pset(
                set().union(*[set(i.bit_bindings)
                              for i in instructions]))  # type: ignore
        else:
            new_qubits = pset(qubits)
        return cls(
            dialect_name=dialect_name,
            instructions=instructions,  # type: ignore
            qubits=new_qubits)  # type: ignore

    def __str__(self):
        return "\n".join([self.dialect_name] + [f"Qubits: {self.qubits}"] +
                         [str(i) for i in self.instructions])


def circuit_conforms_to_dialect(c: Circuit, d: Dialect) -> bool:
    gatedefs_in_circuit = {i.gate_def for i in c.instructions}
    return c.dialect_name == d.name and gatedefs_in_circuit.issubset(
        d.gate_defs)


def circuit_bit_bindings(circuit: Circuit) -> PMap[Tuple[int, int], Set[int]]:
    """
    Given a sequence of instructions, return the complete
    map of bit bindings.  So "H(1), CX(0,1)" would return
    the bit bindings { (1,0):1, (2,0):0, (2,1):1 }
    """
    result: dict = {}
    for i, instruction in enumerate(circuit.instructions):
        for k, v in instruction_bit_bindings_map(instruction).items():
            result[(i, k)] = v
    return pmap(result)


"""
A BitBindingSignature is a type declaration here, but it's
"the set of sets of bit IDs in the circuit instructions which are bound to the
same bit"

So {{(0,0), (1,0)}, {(0,1)}} means that for a two-instruction circuit,
both instruction 0, bit 0 and instruction 1 bit 1 are bound to the same input
bit, and instruction 0, bit 1 is bound to a different bit.

The BitBindingSignature is used to compare two circuits and see if they have
the same basic graph structure
"""
BitBindingSignature = PSet[PSet[Tuple[int, int]]]


def circuit_bit_binding_signature(c: Circuit) -> BitBindingSignature:
    """
    Create the BitBindingSignature of a circuit
    """
    forward_bindings = circuit_bit_bindings(c)
    reverse_bindings = reverse_map(forward_bindings)
    return pset(reverse_bindings.values())


def circuit_parameters_are_fully_bound(c: Circuit) -> bool:
    """
    Whether or not every instruction in the circuit has its parameters
    fully bound
    """
    return all(
        [instruction_parameters_are_fully_bound(i) for i in c.instructions])


def circuit_is_valid_executable(c: Circuit) -> bool:
    """
    Whether not each instruction in the circuit is executable
    """
    return all([instruction_is_valid_executable(i) for i in c.instructions])


def circuit_is_valid_replacement(c: Circuit) -> bool:
    """
    Whether or not each instruction in the circuit is a valid replacement
    """
    return all([instruction_is_valid_replacement(i) for i in c.instructions])


@require(lambda target: circuit_parameters_are_fully_bound(target))
def circuit_pattern_matches_target(pattern: Circuit, target: Circuit) -> bool:
    return ((len(pattern.instructions) == len(target.instructions)) and all([
        instruction_pattern_matches_target(pattern.instructions[i],
                                           target.instructions[i])
        for i in range(len(pattern.instructions))
    ]) and (circuit_bit_binding_signature(pattern)
            == circuit_bit_binding_signature(target)))


def circuit_parameter_map(c: Circuit) -> Mapping[Tuple[int, str], Any]:
    result = {}
    for index, i in enumerate(c.instructions):
        for k, v in i.parameter_bindings.items():
            result[(index, k)] = v
    return pmap(result)


def circuit_parameter_names(c: Circuit) -> PSet[Tuple[int, str]]:
    result = set([])
    for index, i in enumerate(c.instructions):
        for name in i.gate_def.parameter_names:
            result.add((index, name))
    return pset(result)


def circuit_bit_targets(c: Circuit) -> PVector[int]:
    """
    an ordered list of bits addressed by bit bindings, so for
    H(0), CX(0,1) you should get [0,0,1]

    Circuits with the same bit binding signature should have a 
    1:1 correspondence with their circuit_bit_targets
    """
    return pvector(
        itertools.chain.from_iterable([i.bit_bindings for i in c.instructions
                                       ]))  # type: ignore



from qcware_transpile.matching import (TranslationRule, TranslationSet,
                                       trivial_rule, trivial_rules,
                                       untranslatable_instructions,
                                       untranslated_gates, simple_translate,
                                       circuit_is_simply_translatable_by)
from qcware_transpile.dialects import quasar as quasar_dialect, pyzx as pyzx_dialect
import pyzx
from qcware_transpile.circuits import Circuit
from qcware_transpile import TranslationException
from pyrsistent import pset
from icontract.errors import ViolationError
import quasar
from toolz.functoolz import thread_first
from fractions import Fraction
from numpy import pi
from typing import Dict


def to_phase_angle(theta: float) -> Fraction:
    """
    Pyzx seems to evaluate phase angles as a fraction of pi (?), so we must
    translate our quasar angles into that
    """
    return Fraction(theta / pi)


def translation_set():
    """
    Creates a translation set from quasar to qiskit
    """
    trivial_gates = {
        ('H', 'HAD'),
        ('X', 'NOT'),
        ('CZ', 'CZ'),  # ('T', 'T', lambda pm: False),
        ('Rz', 'ZPhase', to_phase_angle),
        ('Z', 'Z'),
        ('CX', 'CNOT'),  #, ('S', 'S')}
        ('Rx', 'XPhase', to_phase_angle),
        ('SWAP', 'SWAP')
    }

    # the T and S gates can be adjoint in pyzx, although in quasar those
    # adjoints are the TT and ST gates respectively
    quasar_d = quasar_dialect.dialect()
    pyzx_d = pyzx_dialect.dialect()
    st_gates = {
        TranslationRule(pattern=Circuit.from_tuples(quasar_d,
                                                    [('T', {}, [0])]),
                        replacement=Circuit.from_tuples(
                            pyzx_d, [('T', {
                                'adjoint': False
                            }, [0])])),
        TranslationRule(pattern=Circuit.from_tuples(quasar_d,
                                                    [('TT', {}, [0])]),
                        replacement=Circuit.from_tuples(
                            pyzx_d, [('T', {
                                'adjoint': True
                            }, [0])])),
        TranslationRule(pattern=Circuit.from_tuples(quasar_d,
                                                    [('S', {}, [0])]),
                        replacement=Circuit.from_tuples(
                            pyzx_d, [('S', {
                                'adjoint': False
                            }, [0])])),
        TranslationRule(pattern=Circuit.from_tuples(quasar_d,
                                                    [('ST', {}, [0])]),
                        replacement=Circuit.from_tuples(
                            pyzx_d, [('T', {
                                'adjoint': True
                            }, [0])])),
    }
    rules = pset().union(trivial_rules(quasar_d, pyzx_d, trivial_gates))
    return TranslationSet(from_dialect=quasar_d,
                          to_dialect=pyzx_d,
                          rules=rules)


target_gatenames = sorted(
    [x.name for x in translation_set().to_dialect.gate_defs])
untranslated = sorted([x.name for x in untranslated_gates(translation_set())])


def audit(c: quasar.Circuit) -> Dict:
    ir_audit = quasar_dialect.audit(c)
    if len(ir_audit.keys()) > 0:
        return ir_audit

    irc = quasar_dialect.native_to_ir(c)
    untranslatable = untranslatable_instructions(irc, translation_set())

    result = {}
    if len(untranslatable) > 0:
        result['untranslatable_instructions'] = untranslatable
    if not quasar.Circuit.test_equivalence(c, c.center()):
        result['circuit_not_centered'] = True # type: ignore
    return result


def native_is_translatable(c: quasar.Circuit):
    """
    A native quasar circuit is translatable to qiskit if it
    is "centered" (ie no leading qubits) and is composed of translatable
    gates
    """
    return len(audit(c)) == 0


def translate(c: quasar.Circuit) -> pyzx.Circuit:
    """
    Native-to-native translation
    """
    if not native_is_translatable(c):
        raise TranslationException(audit(c))
    try:
        return thread_first(c, quasar_dialect.native_to_ir,
                            lambda x: simple_translate(translation_set(), x),
                            pyzx_dialect.ir_to_native)
    except ViolationError:
        raise TranslationException(audit(c))



from qcware_transpile.matching import (TranslationRule, TranslationSet,
                                       trivial_rule, trivial_rules,
                                       translated_gates, untranslated_gates,
                                       simple_translate,
                                       circuit_is_simply_translatable_by,
                                       untranslatable_instructions)
from qcware_transpile.dialects import quasar as quasar_dialect, qiskit as qiskit_dialect
import qiskit
from qcware_transpile import TranslationException
from pyrsistent import pset
from icontract.errors import ViolationError
import quasar
from toolz.functoolz import thread_first


def half_angle(theta):
    return theta / 2.0


def translation_set():
    """
    Creates a translation set from quasar to qiskit
    """
    trivial_gates = {('id', 'I'), ('h', 'H'), ('x', 'X'), ('y', 'Y'),
                     ('z', 'Z'), ('s', 'S'), ('t', 'T'), ('cx', 'CX'),
                     ('cy', 'CY'), ('cz', 'CZ'), ('ccx', 'CCX'),
                     ('swap', 'SWAP'), ('cswap', 'CSWAP'),
                     ('rx', 'Rx', half_angle), ('ry', 'Ry', half_angle),
                     ('rz', 'Rz', half_angle), ('measure', 'I')}  #('u1', 'u1')

    quasar_d = quasar_dialect.dialect()
    qiskit_d = qiskit_dialect.dialect()
    # u1/u2/u3 rules were included but are disabled as IBM is deprecating
    # them at some level.  We currently remove measure rules, but cannot
    # translate reset or barrier rules.
    # we "remove" measure rules by converting them to I gates, which is a hack
    # to allow qiskit circuits with edge bits that are measured.
    rules = pset().union(trivial_rules(
        qiskit_d, quasar_d, trivial_gates))
    return TranslationSet(from_dialect=qiskit_d,
                          to_dialect=quasar_d,
                          rules=rules)


target_gatenames = sorted(
    [x.name for x in translation_set().to_dialect.gate_defs])
untranslated = sorted([x.name for x in untranslated_gates(translation_set())])


def audit(c: qiskit.QuantumCircuit):
    """
    Tries to return a list of issues with the circuit which would
    make it untranslatable
    """
    ir_audit = qiskit_dialect.audit(c)
    if len(ir_audit) > 0:
        return ir_audit

    qdc = qiskit_dialect.native_to_ir(c)
    untranslatable = untranslatable_instructions(qdc, translation_set())

    # circuit_qubits = sorted(list(qdc.qubits))
    # used_qubits = sorted(
    #     list(set().union(*[set(i.bit_bindings) for i in qdc.instructions])))

    result = {}

    # if len(used_qubits) == 0:
    #     result["no_used_qubits"] = True
    # else:
    #     unused_edge_qubits = {
    #         x
    #         for x in circuit_qubits
    #         if (x < used_qubits[0]) or (x > used_qubits[-1])
    #     }
    #     if len(unused_edge_qubits) > 0:
    #         result['unused_edge_qubits'] = unused_edge_qubits

    if len(untranslatable) > 0:
        result['untranslatable_instructions'] = untranslatable
    return result


def basis_gates():
    """The "basis gates" for the quasar backend, according to qiskit
    """
    return list({x.name for x in translated_gates(translation_set())})


def native_is_translatable(c: qiskit.QuantumCircuit, should_transpile=True):
    """
    A native circuit is translatable to quasar if it has no leading or
    following "empty" qubits as currently there is no way to express this in quasar
    """
    c2 = qiskit.compiler.transpile(
        c, basis_gates=basis_gates()) if should_transpile else c.copy()
    return len(audit(c2)) == 0


def translate(c: qiskit.QuantumCircuit,
              should_transpile=True) -> quasar.Circuit:
    """
    Native-to-native translation.  If should_transpile is True,
    attempt to use qiskit to transpile the circuit to the set of 
    gates supported by the translation table.  There exists a chance
    this results in a strange audit with gates which are not in
    the original circuit, but by and large it works
    """
    if should_transpile:
        basis_gates = list(
            {x.name
             for x in translated_gates(translation_set())})
        c2 = qiskit.compiler.transpile(c, basis_gates=basis_gates)
    else:
        c2 = c.copy()
    if not native_is_translatable(c2):
        raise TranslationException(audit(c2))
    try:
        return thread_first(c2, qiskit_dialect.native_to_ir,
                            lambda x: simple_translate(translation_set(), x),
                            lambda x: quasar_dialect.ir_to_native(x, fill_unused_edge_qubits=True))
    except ViolationError:
        raise TranslationException(audit(c2))



"""
Quasar is the toolkit used by QC Ware internally.

It supports fast and accurate simulation and evaluation with
QC Ware's Forge platform
"""
from .quasar_dialect import (dialect, native_to_ir, ir_to_native,
                             native_circuits_are_equivalent, audit)



from qcware_transpile.matching import (TranslationRule, TranslationSet,
                                       trivial_rule, trivial_rules,
                                       untranslated_gates, simple_translate,
                                       circuit_is_simply_translatable_by,
                                       untranslatable_instructions)
from qcware_transpile.dialects import (quasar as quasar_dialect, qiskit as
                                       qiskit_dialect)
from qcware_transpile.circuits import Circuit
from qcware_transpile import TranslationException
from icontract.errors import ViolationError
import quasar
from pyrsistent import pset
import qiskit
from toolz.functoolz import thread_first


def double_angle(theta):
    return 2 * theta


def translation_set():
    """
    Creates a translation set from quasar to qiskit
    """
    trivial_gates = {('I', 'id'), ('H', 'h'), ('X', 'x'),
                     ('Y', 'y'), ('Z', 'z'), ('S', 's'),
                     ('T', 't'), ('CX', 'cx'), ('CY', 'cy'),
                     ('CZ', 'cz'), ('CCX', 'ccx'), ('u1', 'u1'),
                     ('SWAP', 'swap'), ('CSWAP', 'cswap'),
                     ('Rx', 'rx', double_angle),
                     ('Ry', 'ry', double_angle),
                     ('Rz', 'rz', double_angle)}

    quasar_d = quasar_dialect.dialect()
    qiskit_d = qiskit_dialect.dialect()
    other_rules = {
        TranslationRule(
            pattern=Circuit.from_tuples(quasar_d, [('RBS', {}, [0, 1])]),
            replacement=Circuit.from_tuples(
                qiskit_d, [('cx', {}, [1, 0]),
                           ('cry', {
                               'theta': lambda pm: double_angle(pm[
                                   (0, 'theta')])
                           }, [0, 1]), ('cx', {}, [1, 0])]))
    }
    # the U2/U3 rules are disabled for now as they seem to be problematic

    rules = pset().union(trivial_rules(quasar_d, qiskit_d,
                                       trivial_gates)).union(other_rules)
    return TranslationSet(from_dialect=quasar_d,
                          to_dialect=qiskit_d,
                          rules=rules)


target_gatenames = sorted(
    [x.name for x in translation_set().to_dialect.gate_defs])
untranslated = sorted([x.name for x in untranslated_gates(translation_set())])


def audit(c: quasar.Circuit):
    ir_audit = quasar_dialect.audit(c)
    if len(ir_audit.keys()) > 0:
        return ir_audit

    irc = quasar_dialect.native_to_ir(c)
    untranslatable = untranslatable_instructions(irc, translation_set())

    result = {}
    if len(untranslatable) > 0:
        result['untranslatable_instructions'] = untranslatable
    if not quasar.Circuit.test_equivalence(c, c.center()):
        # type notation below since TypedDict was fresh in 3.8
        result['circuit_not_centered'] = True  # type: ignore
    return result


def native_is_translatable(c: quasar.Circuit):
    """
    A native quasar circuit is translatable to qiskit if it
    is "centered" (ie no leading qubits)
    """
    return len(audit(c)) == 0


def translate(c: quasar.Circuit) -> qiskit.QuantumCircuit:
    """
    Native-to-native translation
    """
    if not native_is_translatable(c):
        raise TranslationException(audit(c))
    try:
        return thread_first(c, quasar_dialect.native_to_ir,
                            lambda x: simple_translate(translation_set(), x),
                            qiskit_dialect.ir_to_native)
    except ViolationError:
        raise TranslationException(audit(c))



from qcware_transpile.matching import (TranslationRule, TranslationSet,
                                       trivial_rule, trivial_rules,
                                       untranslated_gates, simple_translate,
                                       circuit_is_simply_translatable_by,
                                       untranslatable_instructions)
from qcware_transpile.dialects import quasar as quasar_dialect, pyzx as pyzx_dialect
import pyzx
from qcware_transpile.circuits import Circuit
from qcware_transpile import TranslationException
from pyrsistent import pset
import quasar
from typing import Dict
from toolz.functoolz import thread_first
from fractions import Fraction
from numpy import pi
from icontract.errors import ViolationError


def from_phase_angle(phase: Fraction) -> float:
    """
    Pyzx seems to evaluate phase angles as a fraction of pi/2, so we must
    translate our quasar angles from that
    """
    return float(phase * pi)


def translation_set():
    """
    Creates a translation set from quasar to qiskit
    """
    trivial_gates = {
        ('HAD', 'H'),
        ('NOT', 'X'),
        ('CZ', 'CZ'),  # ('T', 'T', lambda pm: False),
        ('ZPhase', 'Rz', from_phase_angle),
        ('Z', 'Z'),
        ('CNOT', 'CX'),
        ('XPhase', 'Rx', from_phase_angle),
        ('SWAP', 'SWAP')
    }

    # the T and S gates can be adjoint in pyzx, although in quasar those
    # adjoints are the TT and ST gates respectively
    quasar_d = quasar_dialect.dialect()
    pyzx_d = pyzx_dialect.dialect()
    st_gates = {
        TranslationRule(pattern=Circuit.from_tuples(pyzx_d, [('T', {
            'adjoint': False
        }, [0])]),
                        replacement=Circuit.from_tuples(
                            quasar_d, [('T', {}, [0])])),
        TranslationRule(pattern=Circuit.from_tuples(pyzx_d, [('T', {
            'adjoint': True
        }, [0])]),
                        replacement=Circuit.from_tuples(
                            quasar_d, [('TT', {}, [0])])),
        TranslationRule(pattern=Circuit.from_tuples(pyzx_d, [('S', {
            'adjoint': False
        }, [0])]),
                        replacement=Circuit.from_tuples(
                            quasar_d, [('S', {}, [0])])),
        TranslationRule(pattern=Circuit.from_tuples(pyzx_d, [('S', {
            'adjoint': True
        }, [0])]),
                        replacement=Circuit.from_tuples(
                            quasar_d, [('ST', {}, [0])])),
    }
    rules = pset().union(trivial_rules(pyzx_d, quasar_d,
                                       trivial_gates)).union(st_gates)
    return TranslationSet(from_dialect=pyzx_d,
                          to_dialect=quasar_d,
                          rules=rules)


target_gatenames = sorted(
    [x.name for x in translation_set().to_dialect.gate_defs])
untranslated = sorted([x.name for x in untranslated_gates(translation_set())])


def audit(c: pyzx.Circuit) -> Dict:
    ir_audit = pyzx_dialect.audit(c)
    if len(ir_audit.keys()) > 0:
        return ir_audit

    irc = pyzx_dialect.native_to_ir(c)
    untranslatable = untranslatable_instructions(irc, translation_set())

    result = {}
    if len(untranslatable) > 0:
        result['untranslatable_instructions'] = untranslatable
    return result


def native_is_translatable(c: pyzx.Circuit):
    """
    A native quasar circuit is translatable to qiskit if it
    is "centered" (ie no leading qubits) and is composed of translatable
    gates
    """
    return len(audit(c)) == 0


def translate(c: pyzx.Circuit) -> quasar.Circuit:
    """
    Native-to-native translation
    """
    if not native_is_translatable(c):
        raise TranslationException(audit(c))
    try:
        return thread_first(c, pyzx_dialect.native_to_ir,
                            lambda x: simple_translate(translation_set(), x),
                            quasar_dialect.ir_to_native)
    except ViolationError:
        raise TranslationException(audit(c))



"""
Files for defining gates, gate definitions, and the like
"""
from icontract import require, ensure  # type: ignore
from qcware_transpile.exceptions import TranslationException
import attr
from pyrsistent.typing import PSet
from pyrsistent import pset
from typing import Callable, Optional, Tuple, Union, Iterable
from qcware_transpile.helpers import map_seq_to_seq_unique
from qcware_transpile.instructions import Instruction, remapped_instruction, audit_instruction_for_executable
from qcware_transpile.gates import Dialect, GateDef
from qcware_transpile.circuits import (
    Circuit, circuit_bit_targets, circuit_is_valid_replacement,
    circuit_is_valid_executable, circuit_pattern_matches_target,
    circuit_parameter_map, circuit_conforms_to_dialect,
    circuit_parameter_names)


@attr.s(frozen=True)
class TranslationRule(object):
    """
    A translation: a map from one circuit to another.  The replacement
    must be a fully-bound circuit (all parameters for all instructions
    must be bound) with the same set of input bits as the pattern
    """
    pattern = attr.ib(type=Circuit)
    replacement = attr.ib(type=Circuit)

    @replacement.validator
    def check_replacement(self, attribute, value):
        # replacement must address the same bits as pattern
        pattern_bits = circuit_bit_targets(self.pattern)
        replacement_bits = circuit_bit_targets(self.replacement)
        if len(replacement_bits) > 0 and set(pattern_bits) != set(
                replacement_bits):
            raise ValueError(
                f"pattern and replacement must target the same set of bits ({pattern_bits}!={replacement_bits})"
            )
        # replacement values must either be values or keys into pattern
        if not circuit_is_valid_replacement(self.replacement):
            raise ValueError(
                "Replacement circuit must be a valid replacement (fully bound with replacement-style parameter values)"
            )
        pattern_parameters = circuit_parameter_names(self.pattern)
        pm_replacement = circuit_parameter_map(self.replacement)
        pattern_keys = [
            x for x in pm_replacement.values() if isinstance(x, tuple)
        ]
        if not set(pattern_keys).issubset(pattern_parameters):
            raise ValueError(
                "Some replacement values reference tuples not in the source pattern"
            )

    def __str__(self):
        return "\n->\n".join([str(self.pattern), str(self.replacement)])


@require(lambda a, b: len(a.qubit_ids) == len(b.qubit_ids))
@require(lambda a, b: len(a.parameter_names) == len(b.parameter_names) and len(
    a.parameter_names) <= 1)
@require(lambda dialect_a, a, dialect_b, b: a in dialect_a.gate_defs and b in
         dialect_b.gate_defs)
def trivial_rule(dialect_a: Dialect,
                 a: GateDef,
                 dialect_b: Dialect,
                 b: GateDef,
                 f: Optional[Callable] = None) -> TranslationRule:
    """
    Create a trivial translation rule between two gate definitions,
    optionally with a translation lambda for the single parameter.  
    Unlike the target translation function, which takes the single
    parameter of a parameter map (which one would dereference for 
    actual arguments, ie "lambda pm: pm[(0,'theta')]*2") this takes
    a single value and returns a modified value, presuming that there
    single parameter shared by both gates is the same type

    This requires two gatedefs with the same number of qubits and
    up to one parameter.  Beyond that you'll have to be a bit more clever.
    """
    num_qubits = len(a.qubit_ids)
    pattern_instruction = Instruction(gate_def=a,
                                      bit_bindings=range(num_qubits),
                                      parameter_bindings={})
    if len(a.parameter_names) == 1:
        a_parameter_name = list(a.parameter_names)[0]
        b_parameter_name = list(b.parameter_names)[0]

        b_parameter_target = (0, a_parameter_name)
        b_parameter_value = b_parameter_target if f is None else lambda pm: f(
            pm[b_parameter_target])
        parameter_bindings = {b_parameter_name: b_parameter_value}
    else:
        parameter_bindings = {}
    target_instruction = Instruction(
        gate_def=b,
        bit_bindings=range(num_qubits),
        parameter_bindings=parameter_bindings)  # type: ignore
    return TranslationRule(
        pattern=Circuit.from_instructions(dialect_a.name,
                                          [pattern_instruction]),
        replacement=Circuit.from_instructions(dialect_b.name,
                                              [target_instruction]))


def trivial_rules(
    dialect_a: Dialect, dialect_b: Dialect,
    name_tuples: Iterable[Union[Tuple[str, str], Tuple[str, str, Callable]]]
) -> PSet[TranslationRule]:
    """
    Create a set of trivial rules without any parameter translation
    by pairs of gate names
    """
    return pset({
        trivial_rule(dialect_a, dialect_a.gate_named(t[0]), dialect_b,
                     dialect_b.gate_named(t[1]),
                     None if len(t) == 2 else t[2])  # type: ignore
        for t in name_tuples
    })


@require(lambda translation, target: circuit_pattern_matches_target(
    translation.pattern, target))
@require(lambda target: circuit_is_valid_executable(target))
def translation_replace_circuit(translation: TranslationRule,
                                target: Circuit) -> Circuit:
    pattern_bits = list(circuit_bit_targets(translation.pattern))
    target_bits = list(circuit_bit_targets(target))
    # okay, at this point the pattern and target bits could have redundant pairs,
    # but should not have mixed pairs, ie [1,2,1,0] [3,2,3,1] is ok, but
    # [1,2,1,0] [3,2,2,0] is not ok because 1 maps to 2 different things
    bitpairs = set(zip(pattern_bits, target_bits))
    pattern_bits = [x[0] for x in bitpairs]
    target_bits = [x[1] for x in bitpairs]
    qubit_map = map_seq_to_seq_unique(pattern_bits, target_bits)
    parameter_map = circuit_parameter_map(target)
    replacement_instructions = [
        remapped_instruction(qubit_map, parameter_map, i)
        for i in translation.replacement.instructions  # was target
    ]
    # hmm, ignoring the following because it raised a curious
    # error: Argument "instructions" to "Circuit" has incompatible
    # type "List[Instruction]"; expected "Iterable[T]"
    return Circuit.from_instructions(
        dialect_name=translation.replacement.dialect_name,
        instructions=replacement_instructions)  # type: ignore


@attr.s
class TranslationSet(object):
    """
    A TranslationSet is a translation between one Dialect and another.
    """
    from_dialect = attr.ib(type=Dialect)
    to_dialect = attr.ib(type=Dialect)
    rules = attr.ib(type=PSet[TranslationRule], converter=pset)

    def __str__(self):
        return "\n  ".join(
            ["->".join([self.from_dialect.name, self.to_dialect.name])] +
            [str(rule) for rule in self.rules])


def matching_rules(ts: TranslationSet, c: Circuit) -> PSet[TranslationRule]:
    """
    The set of rules from the translation set which match the circuit.
    Right now this is not efficient; it could be assisted by maps based on
    first gate name in the rule, etc.
    """
    return pset(
        {r
         for r in ts.rules if circuit_pattern_matches_target(r.pattern, c)})


def circuit_is_simply_translatable_by(c: Circuit, ts: TranslationSet) -> bool:
    """
    A simple translation is one where each instruction in the circuit is
    addressed by a single-instruction pattern in the translation set
    """
    return len(untranslatable_instructions(c, ts)) == 0


def untranslatable_instructions(c: Circuit,
                                ts: TranslationSet) -> PSet[Instruction]:
    subcircuits = [
        Circuit.from_instructions(c.dialect_name, [i]) for i in c.instructions
    ]  # type: ignore
    return pset({x for x in subcircuits if len(matching_rules(ts, x)) == 0})


@require(lambda ts, target: target.dialect_name == ts.from_dialect.name)
@require(lambda ts, target: len(matching_rules(ts, target)) > 0)
def translationset_replace_circuit(ts: TranslationSet,
                                   target: Circuit) -> Circuit:
    """
    Compute a replacement circuit by the given translation set
    """
    # choose an arbitrary matching rule
    rule = list(matching_rules(ts, target))[0]
    if not circuit_is_valid_executable(target):
        audit = {"nonexecutable_target": target}
        instruction_audits = [
            audit_instruction_for_executable(i) for i in target.instructions
        ]
        audit[
            'nonexecutable_instructions'] = instruction_audits  # type: ignore
        raise TranslationException(audit=audit)
    result = translation_replace_circuit(rule, target)
    return result


@require(lambda ts, c: circuit_conforms_to_dialect(c, ts.from_dialect))
@require(lambda ts, c: circuit_is_simply_translatable_by(c, ts))
@ensure(lambda result, ts: circuit_conforms_to_dialect(result, ts.to_dialect))
def simple_translate(ts: TranslationSet, c: Circuit) -> Circuit:
    """
    Given a simple translation set ts, breaks c into subcircuits
    of length 1 (ie individual instructions), translates each instruction,
    and assembles the result into a circuit belonging to the target
    dialect
    """
    subcircuits = [
        Circuit.from_instructions(c.dialect_name, [i]) for i in c.instructions
    ]  # type: ignore
    new_instructions = [
        instruction for sub in subcircuits
        for instruction in translationset_replace_circuit(ts, sub).instructions
    ]
    return Circuit.from_instructions(dialect_name=ts.to_dialect.name,
                                instructions=new_instructions,
                                qubits = c.qubits) # type: ignore


def translated_gates(tset: TranslationSet) -> PSet[GateDef]:
    """
    Given a translation set, give a list of GateDefs which
    are translated.  This only matches rules in the TranslationSet
    which are one instruction long and have no bound
    parameters!
    """
    def is_simple_pattern(x: Circuit):
        return len(x.instructions) == 1 and len(
            x.instructions[0].parameter_bindings) == 0

    return pset({x.pattern.instructions[0].gate_def for x in tset.rules})


def untranslated_gates(tset: TranslationSet) -> PSet[GateDef]:
    """
    Returns the set of GateDefs in the TranslationSet's 'from_dialect'
    which do not have a corresponding simple rule in the TranslationSet's
    rules
    """
    return pset(tset.from_dialect.gate_defs - translated_gates(tset))



from quasar.circuit import Gate  # type: ignore
from quasar.circuit import Circuit as QuasarCircuit  # type: ignore
from pyrsistent import pset
from pyrsistent.typing import PSet
from typing import Callable, Any, Tuple, Generator, Dict
# not using relative imports here for the moment to simplify emacs
# send-to-buffer issues; it should arguably be set back to relative
# imports at some point
from qcware_transpile.gates import GateDef, Dialect
from qcware_transpile.circuits import Circuit
from qcware_transpile.instructions import Instruction
from qcware_transpile.helpers import map_seq_to_seq_unique
from inspect import signature
from icontract import require  # type: ignore
import functools

__dialect_name__ = "quasar"


def is_builtin(thing) -> bool:
    """
    Whether or not something is a Python builtin; used to filter
    out things in the Gate namespace which represent a Gate or the
    function used to create a gate.
    """
    return type(thing).__name__ in dir(__builtins__) or type(
        thing).__name__ == "builtin_function_or_method"


def represents_gate(f) -> bool:
    """
    As expected: whether or not a thing represents a gate
    (either by being a Gate or by being a callable with
    fully-populated defaults which returns a Gate object)
    """
    if isinstance(f, Gate):
        return True
    if (not callable(f)) or is_builtin(f):
        return False
    try:
        sig = signature(f)
    # this is genuinely strange.  When this buffer is sent to a
    # python interpreter, quasar_gatenames_full, etc does just fine;
    # when it's imported, it blows up without the following because
    # it raises "ValueError for item <class 'type'>
    except ValueError:
        return False
    has_all_defaults = all([
        param.default is not param.empty for param in sig.parameters.values()
    ])
    if not has_all_defaults:
        return False
    f_result = f()
    if not isinstance(f_result, Gate):
        return False
    return True


def gatedef_from_gate(name: str, g: Gate) -> GateDef:
    """
    Create a gatedef from a gate object.  We use a given
    name instead of g.name because gate "names" in quasar
    don't always correspond to the Python names (for example,
    Gate.ST has name = 'S^+', but for our dialect we want
    the mapping to Python to be as simple as possible)
    """
    name = name
    parameter_names = {k for k in g.parameters.keys()}
    num_qubits = g.nqubit
    return GateDef(name=name,
                   parameter_names=parameter_names,
                   qubit_ids=num_qubits)


def gatedef_from_gatefun(name: str, g: Callable) -> GateDef:
    """
    Creates a gatedef from a function that returns a gate;
    see above for reasons we have an explicit name
    """
    return gatedef_from_gate(name, g())


@require(lambda thing: represents_gate(thing))
def gatedef_from_gatething(name: str, thing: Any) -> GateDef:
    if isinstance(thing, Gate):
        return gatedef_from_gate(name, thing)
    elif callable(thing):
        return gatedef_from_gatefun(name, thing)
    else:
        assert False


@require(lambda thing: represents_gate(thing))
def gate_name_property(thing: Any) -> str:
    """
    Return the .name property of a Gate (or Gate
    returned by a callable function)
    """
    if isinstance(thing, Gate):
        return thing.name
    elif callable(thing):
        g = thing()
        assert (isinstance(g, Gate))
        return g.name
    else:
        assert False
        return None


@functools.lru_cache(1)
def quasar_names_full() -> PSet[str]:
    return pset(dir(Gate))


@functools.lru_cache(1)
def quasar_gatenames_full() -> PSet[str]:
    """
    The names of verything in the Gate namespace
    """
    result = pset(
        {name
         for name in dir(Gate) if represents_gate(getattr(Gate, name))})
    return result


@functools.lru_cache(1)
def quasar_gatethings_full() -> PSet:
    """
    All the things in the Gate namespace which represent a gate
    """
    result = pset({
        getattr(Gate, name)
        for name in quasar_gatenames_full()
        if represents_gate(getattr(Gate, name))
    })
    return result


@functools.lru_cache(1)
def dialect() -> Dialect:
    """
    Programmatically create the Quasar dialect
    """
    gatedefs = {
        gatedef_from_gatething(name, getattr(Gate, name))
        for name in quasar_gatenames_full()
    }
    return Dialect(name="quasar", gate_defs=gatedefs)  # type: ignore


@functools.lru_cache(1)
def name_property_to_namespace_translation_table():
    """
    Okay... names in the Gate namespace for gates ("ST") don't always translate
    to the gate's .name property (in the ST case, "S^+").  So this is a translation
    table from the gate's name property to the name in the Gate namespace
    """
    namespace_names = list(quasar_gatenames_full())
    name_property = [
        gate_name_property(getattr(Gate, name)) for name in namespace_names
    ]
    return map_seq_to_seq_unique(name_property, namespace_names)


def native_instructions(
        qc: QuasarCircuit) -> Generator[Tuple[Gate, Tuple[int]], None, None]:
    """
    Iterates through the circuit yielding tuples of gate and 
    qubits.
    """
    for key, gate in qc.gates.items():
        yield gate, key[1]


def ir_instruction_from_native(gate: Gate, qubits: Tuple[int]) -> Instruction:
    """
    Direct conversion of one gate/qubits tuple into an IR "instruction"
    """
    ntt = name_property_to_namespace_translation_table()
    return Instruction(gate_def=gatedef_from_gatething(ntt[gate.name], gate),
                       bit_bindings=qubits,
                       parameter_bindings=gate.parameters)


def native_to_ir(qc: QuasarCircuit) -> Circuit:
    """
    Return a transpile-style Circuit object from a quasar Circuit object
    """
    instructions = list(
        ir_instruction_from_native(x[0], x[1])
        for x in native_instructions(qc))
    return Circuit.from_instructions(dialect_name=__dialect_name__,
                                     instructions=instructions)  # type: ignore


def quasar_gate_from_instruction(i: Instruction) -> Gate:
    """
    Create a quasar Gate object from an instruction
    """
    g = getattr(Gate, i.gate_def.name)
    # if g is a Gate, it shouldn't have parameters; if it's a callable,
    # it should.
    if len(i.parameter_bindings.keys()) > 0:
        g = g(**i.parameter_bindings)
    return g


def ir_to_native(c: Circuit, fill_unused_edge_qubits=False) -> QuasarCircuit:
    result = QuasarCircuit()
    for instruction in c.instructions:
        g = quasar_gate_from_instruction(instruction)
        result.add_gate(g, tuple(instruction.bit_bindings))
    # by default, quasar silently ignores unused edge qubits (qubits
    # "above" or "below" instructions that have no instructions of their
    # own).  This can cause disparity with toolkits that do not
    # share this behaviour.  It unfortunately extends the current idea
    # that qubit indices are integers
    if fill_unused_edge_qubits == True:
        # find the unused edge qubits.  We can use quasar's min_qubit and max_qubit
        # combined with the list of qubits in the Circuit object.
        # The Circuit's qubits is a set
        qubits_to_fill = { qb for qb in c.qubits if (qb < result.min_qubit) or (qb > result.max_qubit) }
        # on each unused edge qubits, add an identity gate
        for qb in qubits_to_fill:
            result.I(qb)
    return result 


def native_circuits_are_equivalent(c1: QuasarCircuit,
                                   c2: QuasarCircuit) -> bool:
    """
    Our own definition of whether two quasar circuits are equivalent;
    used to test
    """
    return QuasarCircuit.test_equivalence(c1, c2)


def audit(c: QuasarCircuit) -> Dict:
    """
    Right now, anything expressible in Quasar *should*
    be expressible in the IR, so this is a dummy result.
    """
    return {}



"""
Some helper functions
"""
from typing import Sequence, Mapping
from icontract import require  # type: ignore
from pyrsistent import pmap, pset
from typing import TypeVar, Dict, Iterable, Callable
from pyrsistent.typing import PMap, PSet

A = TypeVar('A')
B = TypeVar('B')
T = TypeVar('T')


@require(lambda s1, s2: len(s1) == len(s2))
def map_seq_to_seq(s1: Sequence[A], s2: Sequence[B]) -> PMap[A, PSet[B]]:
    """
    Given two sequences of equal lengths, map each
    element of s1 to the set of the corresponding element in s2,
    such that domain(result) == set(s1); return the mapping
    as a pmap

    If s1 has multiple entries with equal values, the 
    results for f(x) would be {y,z...}; in other words,
    mapping [1,2,1] onto [3,4,5] would result in
    { 1: {3,5}, 2: {4} }
    """
    result: dict = {}
    for i in range(len(s1)):
        x = s1[i]
        y = s2[i]
        if x in result:
            result[x] = result[x].add(y)
        else:
            result[x] = pset([y])
    # we now have a mapping of Any to Set[Any], but since
    # we'll be reversing this later, we must map all values to PSet[Any]
    return pmap(result)


@require(lambda s1, s2: len(s1) == len(s2))
@require(lambda s1: len(set(s1)) == len(s1))
def map_seq_to_seq_unique(s1: Sequence[A], s2: Sequence[B]) -> PMap[A, B]:
    """
    Much like map_seq_to_seq, in this case provide a direct
    mapping between the sequences (s1 and S2 must have unique values and
    no mapping from s1 to s2 should be multiply defined)

    The implementation is pretty darn obvious; this definition is mostly for
    the contracts and typing.
    """
    result = pmap(dict(zip(s1, s2)))
    return result


def prepend_index_to_domain(index: int, f: Mapping):
    """
    Transforms the keys of a mapping into a tuple with
    the index as the first element; in other words,
    if f = { 1: 2, 3: 4 }, prepend_index_to_domain(3,f)
    would return { (3,1): 2, (3,3): 4 }
    """
    return pmap({(index, k): v for k, v in f.items()})


def reverse_map(f: Mapping[A, B]) -> PMap[B, PSet[A]]:
    """
    Reverses a mapping such that the values of f
    are mapped to the sets of keys in f corresponding 
    to that value
    """
    result: Dict[B, PSet[A]] = {}
    for k, v in f.items():
        if v in result:
            result[v] = result[v].add(k)
        else:
            result[v] = pset([k])
    return pmap(result)


def exists_in(i: Iterable[T], pred: Callable[[T], bool]) -> bool:
    """
    Checks for any item in i such that pred(i) is true
    """
    for x in i:
        if pred(x):
            return True
    return False



from pyrsistent import pvector, pset, pmap
from pyrsistent.typing import PSet, PMap
from icontract import require  # type: ignore
from typing import Union, Sequence, Set
import attr


def _qubit_ids(qubits: Union[int, Sequence[int]]):
    return pvector(range(qubits)) if isinstance(qubits,
                                                int) else pvector(qubits)


@attr.s(frozen=True)
class GateDef(object):
    """
    A gate definition, consisting only of a name, a set of names
    for parameters, and an ordered collection of integer qubit IDs.

    GateDefs can also support variable numbers of qubit args if 
    varargs == true.  This is not currently exploited
    """
    name = attr.ib(type=str)
    parameter_names = attr.ib(type=Set[str], converter=pset)
    qubit_ids = attr.ib(type=Sequence[int], converter=_qubit_ids)
    has_varargs = attr.ib(type=bool, default=False)

    def __str__(self):
        return "".join([self.name, "("] +
                       [",".join([s
                                  for s in self.parameter_names])] + ["), ("] +
                       [",".join([str(i) for i in self.qubit_ids])] + [")"])


@attr.s(frozen=True)
class Dialect(object):
    """
    A "dialect" -- essentially just a named set of gate definitions
    """
    name = attr.ib(type=str)
    gate_defs = attr.ib(type=PSet[GateDef], converter=pset)
    gate_map = attr.ib(type=PMap[str, GateDef], init=False)

    def __attrs_post_init__(self):
        # build an index by gate name to speed things up a bit
        gate_map: PMap[str,
                       GateDef] = pmap({g.name: g
                                        for g in self.gate_defs})
        object.__setattr__(self, "gate_map", gate_map)

    def __str__(self):
        return "\n  ".join([self.name] + [str(g) for g in self.gate_defs])

    def has_gate_named(self, name: str) -> bool:
        return name in self.gate_map

    @require(lambda self, name: self.has_gate_named(name))
    def gate_named(self, name: str) -> GateDef:
        return self.gate_map[name]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['qcware_transpile',
 'qcware_transpile.dialects',
 'qcware_transpile.dialects.pyzx',
 'qcware_transpile.dialects.qiskit',
 'qcware_transpile.dialects.quasar',
 'qcware_transpile.translations',
 'qcware_transpile.translations.pyzx',
 'qcware_transpile.translations.pyzx.to_quasar',
 'qcware_transpile.translations.qiskit',
 'qcware_transpile.translations.qiskit.to_quasar',
 'qcware_transpile.translations.quasar',
 'qcware_transpile.translations.quasar.to_pyzx',
 'qcware_transpile.translations.quasar.to_qiskit']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=20.3.0,<21.0.0',
 'icontract>=2.4.1,<3.0.0',
 'pyrsistent>=0.17.3,<0.18.0',
 'toolz>=0.11.1,<0.12.0']

extras_require = \
{'pyzx': ['pyzx>=0.6.3,<0.7.0'],
 'qcware-quasar': ['qcware-quasar>=1.0.2,<2.0.0'],
 'qiskit': ['qiskit-aer>=0.7.2,<0.8.0']}

setup_kwargs = {
    'name': 'qcware-transpile',
    'version': '0.1.1a5',
    'description': 'A quantum circuit transpilation framework',
    'long_description': None,
    'author': 'Vic Putz',
    'author_email': 'vic.putz@qcware.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
Modules and functions for the transpilation of circuits
"""
try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata

__version__ = importlib_metadata.version(__name__)

from .exceptions import TranslationException



"""
Qiskit is IBM's rather complex toolkit, including many features such as
decomposition to primitive gates and transpilation.
"""

from .qiskit_dialect import (dialect, native_to_ir, ir_to_native,
                             native_circuits_are_equivalent, audit)
